@@ -1,6 +1,6 @@
 /* channel_commands.h
  *
- * $Id: channel_commands.h,v 1.6 1999/04/19 19:35:04 nisse Exp $
+ * $Id: channel_commands.h,v 1.8 1999/06/11 20:40:25 nisse Exp $
  */
 
 /* lsh, an implementation of the ssh protocol
@@ -73,5 +73,22 @@ int do_channel_open_command(struct command *s,
 int do_channel_request_command(struct command *s,
 			       struct lsh_object *x,
 			       struct command_continuation *c);
-     
+
+/* GABA:
+   (class
+     (name global_request_command)
+     (super command)
+     (vars
+       (format_request method "struct lsh_string *"
+                       "struct ssh_connection *connection"
+		       "struct command_continuation **c")))
+*/
+
+#define FORMAT_GLOBAL_REQUEST(r, conn, c) \
+((r)->format_request((r), (conn), (c)))
+
+int do_channel_global_command(struct command *s,
+			      struct lsh_object *x,
+			      struct command_continuation *c);
+
 #endif /* LSH_CHANNEL_COMMANDS_H_INCLUDED */
